from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS EC2 resources
    Args:
        hub:
        ctx:
        resource_id: aws resource id
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    tags_to_add = list()
    tags_to_remove = list()
    old_tags_map = {tag.get("Key"): tag for tag in old_tags}
    if not isinstance(new_tags, type(None)):
        for tag in new_tags:
            if tag.get("Key") in old_tags_map:
                if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                    del old_tags_map[tag.get("Key")]
                else:
                    tags_to_add.append(tag)
            else:
                tags_to_add.append(tag)
        tags_to_remove = [{"Key": tag.get("Key")} for tag in old_tags_map.values()]
    result = dict(comment="", result=True, ret=None)
    if tags_to_remove:
        delete_ret = await hub.exec.boto3.client.ec2.delete_tags(
            ctx, Resources=[resource_id], Tags=tags_to_remove
        )
        if not delete_ret["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
        result["ret"] = delete_ret["ret"]
    if tags_to_add:
        add_ret = await hub.exec.boto3.client.ec2.create_tags(
            ctx, Resources=[resource_id], Tags=tags_to_add
        )
        if not add_ret["result"]:
            result["comment"] = add_ret["comment"]
            result["result"] = False
            return result
        result["ret"] = add_ret["ret"]
    result["comment"] = f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]"
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_vpc
hub.exec.boto3.client.ec2.delete_vpc
hub.exec.boto3.client.ec2.describe_vpcs
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
hub.tool.boto3.resource.exec(resource, associate_dhcp_options, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_network_acl, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route_table, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_security_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, describe_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, disable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, enable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, modify_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, request_vpc_peering_connection, *args, **kwargs)
"""
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    cidr_block_association_set: List = None,
    ipv6_cidr_block_association_set: List = None,
    instance_tenancy: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a VPC with the specified IPv4 CIDR block. The smallest VPC you can create uses a /28 netmask (16 IPv4
    addresses), and the largest uses a /16 netmask (65,536 IPv4 addresses). For more information about how large to
    make your VPC, see Your VPC and subnets in the Amazon Virtual Private Cloud User Guide. You can optionally
    request an IPv6 CIDR block for the VPC. You can request an Amazon-provided IPv6 CIDR block from Amazon's pool of
    IPv6 addresses, or an IPv6 CIDR block from an IPv6 address pool that you provisioned through bring your own IP
    addresses (BYOIP). By default, each instance you launch in the VPC has the default DHCP options, which include
    only a default DNS server that we provide (AmazonProvidedDNS). For more information, see DHCP options sets in
    the Amazon Virtual Private Cloud User Guide. You can specify the instance tenancy value for the VPC when you
    create it. You can't change this value for the VPC after you create it. For more information, see Dedicated
    Instances in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name(Text): A name, ID to identify the resource.
        cidr_block_association_set(List, optional): Information about the IPv4 CIDR blocks associated with the VPC.
            Defaults to None.
            * CidrBlock (string) -- An IPv4 CIDR block to associate with the VPC.
            * Ipv4IpamPoolId (string) -- Associate a CIDR allocated from an IPv4 IPAM pool to a VPC.
            * Ipv4NetmaskLength (integer) -- The netmask length of the IPv4 CIDR you would like to associate from an Amazon VPC IP Address Manager (IPAM) pool.
        ipv6_cidr_block_association_set(List, optional): Information about the IPv6 CIDR blocks associated with the VPC.
            Defaults to None.
            * Ipv6CidrBlock (string) -- An IPv6 CIDR block from the IPv6 address pool. You must also specify Ipv6Pool in the request.
            * Ipv6IpamPoolId (string) -- Associates a CIDR allocated from an IPv6 IPAM pool to a VPC.
            * Ipv6NetmaskLength (integer) -- The netmask length of the IPv6 CIDR you would like to associate from an
             Amazon VPC IP Address Manager (IPAM) pool.
            * Ipv6CidrBlockNetworkBorderGroup (string) -- The name of the location from which we advertise the IPV6 CIDR
             block. Use this parameter to limit the CIDR block to this location. You must set AmazonProvidedIpv6CidrBlock
              to true to use this parameter. You can have one IPv6 CIDR block association per network border group.
            * AmazonProvidedIpv6CidrBlock (boolean) -- Requests an Amazon-provided IPv6 CIDR block with a /56 prefix
             length for the VPC. You cannot specify the range of IPv6 addresses, or the size of the CIDR block.
        instance_tenancy(Text, optional): The tenancy options for instances launched into the VPC. For default, instances are launched
            with shared tenancy by default. You can launch instances with any tenancy into a shared tenancy
            VPC. For dedicated, instances are launched as dedicated tenancy instances by default. You can
            only launch instances with a tenancy of dedicated or host into a dedicated tenancy VPC.
            Important: The host value cannot be used with this parameter. Use the default or dedicated
            values only. Default: default. Defaults to None.
        tags(List, optional): The tags to assign to the VPC. Defaults to None.
            * Key (string) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [vpc-resource-id]:
          aws.ec2.vpc.present:
          - cidr_block_association_set:
            - CidrBlock: 'string'
              Ipv4IpamPoolId: 'string'
              Ipv4NetmaskLength: 'integer'
          - ipv6_cidr_block_association_set:
            - Ipv6CidrBlock: 'string'
              Ipv6IpamPoolId: 'string'
              Ipv6NetmaskLength: 'integer'
              Ipv6CidrBlockNetworkBorderGroup: 'string'
              AmazonProvidedIpv6CidrBlock: True|False
          - instance_tenancy: 'default'|'dedicated'|'host'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpc-01234672f3336db8:
              aws.ec2.vpc.present:
              - cidr_block_association_set:
                - CidrBlock: 10.1.150.0/28
              - instance_tenancy: default
              - tags:
                - Key: Name
                  Value: vpc-name
                - Key: vpc-tag-key-2
                  Value: vpc-tag-value-2
    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
    before = await hub.tool.boto3.resource.describe(resource)
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.ec2.instance {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.ec2.instance {name}"
            result["result"] = True
        return result

    if before:
        try:
            result["old_state"] = before
            # Update cidr blocks
            update_ret = await hub.exec.aws.ec2.vpc.update_cidr_blocks(
                ctx=ctx,
                vpc_id=before.get("VpcId"),
                old_ipv4_cidr_blocks=before.get("CidrBlockAssociationSet", list()),
                old_ipv6_cidr_blocks=before.get("Ipv6CidrBlockAssociationSet", list()),
                new_ipv4_cidr_blocks=cidr_block_association_set,
                new_ipv6_cidr_blocks=ipv6_cidr_block_association_set,
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
            if result["result"] and (tags is not None):
                # Update tags
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=before.get("VpcId"),
                    old_tags=before.get("Tags"),
                    new_tags=tags,
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
        if result["result"]:
            result["comment"] = f"'Update {name}'"
    else:
        cidr_request_payload = dict()
        # Since boto3 only allows one cidr association when creating a vpc, we use the first cidr associations
        # during vpc creation, associate the rest after creation.
        if cidr_block_association_set:
            cidr_request_payload = (
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    cidr_block_association_set[0], "ipv4"
                )
            )
            cidr_block_association_set.pop(0)
        elif ipv6_cidr_block_association_set:
            cidr_request_payload = (
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    ipv6_cidr_block_association_set[0], "ipv6"
                )
            )
            ipv6_cidr_block_association_set.pop(0)
        try:
            ret = await hub.exec.boto3.client.ec2.create_vpc(
                ctx,
                **cidr_request_payload,
                **{
                    "InstanceTenancy": instance_tenancy,
                    "TagSpecifications": [{"ResourceType": "vpc", "Tags": tags}],
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
            resource._id = ret["ret"]["Vpc"]["VpcId"]
            # Associate the rest cidr associations
            update_ret = await hub.exec.aws.ec2.vpc.update_cidr_blocks(
                ctx=ctx,
                vpc_id=resource._id,
                old_ipv4_cidr_blocks=list(),
                old_ipv6_cidr_blocks=list(),
                new_ipv4_cidr_blocks=cidr_block_association_set,
                new_ipv6_cidr_blocks=ipv6_cidr_block_association_set,
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"]
                result["result"] = False

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["new_state"] = after
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC. You must detach or delete all gateways and resources that are associated with the VPC
    before you can delete it. For example, you must terminate all instances running in the VPC, delete all security
    groups associated with the VPC (except the default one), delete all route tables associated with the VPC (except
    the default one), and so on.

    Args:
        name(Text): The ID of the VPC.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpc-01234672f3336db8:
              aws.ec2.vpc.absent:
                - name: value
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.ec2.vpc {name}"
        return result
    else:
        result["old_state"] = before
        try:
            ret = await hub.exec.boto3.client.ec2.delete_vpc(ctx, **{"VpcId": name})
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe VPCs {ret['comment']}")
        return {}

    for resource in ret["ret"]["Vpcs"]:
        resource_id = resource.get("VpcId")
        describe_parameters = OrderedDict(
            {"InstanceTenancy": "instance_tenancy", "Tags": "tags"}
        )
        resource_translated = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if resource.get(parameter_old_key) is not None:
                resource_translated.append(
                    {parameter_new_key: resource.get(parameter_old_key)}
                )
        # The following code block is to make sure the describe function only outputs the associated cidr blocks and
        # dis-regard the disassociated cidr blocks.
        if resource.get("CidrBlockAssociationSet"):
            ipv4_cidr_block_association_set = list()
            for cidr_block in resource.get("CidrBlockAssociationSet"):
                if cidr_block.get("CidrBlockState").get("State") == "associated":
                    ipv4_cidr_block_association_set.append(cidr_block)
            resource_translated.append(
                {"cidr_block_association_set": ipv4_cidr_block_association_set}
            )
        if resource.get("Ipv6CidrBlockAssociationSet"):
            ipv6_cidr_block_association_set = list()
            for cidr_block in resource.get("Ipv6CidrBlockAssociationSet"):
                if cidr_block.get("Ipv6CidrBlockState").get("State") == "associated":
                    if "NetworkBorderGroup" in cidr_block:
                        # Translate describe output to the correct present input format
                        ipv6_cidr_block_network_border_group = cidr_block.pop(
                            "NetworkBorderGroup"
                        )
                        cidr_block[
                            "Ipv6CidrBlockNetworkBorderGroup"
                        ] = ipv6_cidr_block_network_border_group
                    ipv6_cidr_block_association_set.append(cidr_block)
            resource_translated.append(
                {"ipv6_cidr_block_association_set": ipv6_cidr_block_association_set}
            )

        result[resource_id] = {"aws.ec2.vpc.present": resource_translated}
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.rds.create_db_subnet_group
hub.exec.boto3.client.rds.delete_db_subnet_group
hub.exec.boto3.client.rds.describe_db_subnet_groups
hub.exec.boto3.client.rds.modify_db_subnet_group
"""
import copy
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    db_subnet_group_description: str,
    subnets: List = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new DB subnet group. DB subnet groups must contain at least one subnet in at least two AZs in the
    Amazon Web Services Region.

    Args:
        name(Text, mandatory): A name, ID to identify the resource.
        db_subnet_group_description(TEXT, mandatory): DB Subnet Group description
        subnets(List, mandatory): The EC2 Subnet IDs for the DB subnet group
        tags:(List, optional): Tags to assign to the DB subnet group.

    Request Syntax:
        [db-subnet-group-name]:
          aws.rds.db_subnet_group.absent:
            - db_subnet_group_description: 'string'
            - subnets: 'list'
            - tags:
              - Key: 'string'
                Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            new-db-subnet-group:
                aws.rds.db_subnet_group.absent:
                  - db_subnet_group_description: "should not be blank"
                  - subnets:
                    - subnet-03e29c119d78899e1
                    - subnet-04dd5bc5aa08cbd89
                    - subnet-0754289517e31d331
                  - tags:
                    - Key: name
                      Value: new-db-subnet-group
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
        ctx, DBSubnetGroupName=name
    )
    old_db_subnet_group = {}
    is_updated = False
    is_tags_updated = False
    is_new = False
    db_subnet_group_arn = None
    old_tag_list = None
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.rds.db_subnet_group {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.rds.db_subnet_group {name}"
            result["result"] = True
        return result
    if before:
        old_db_subnet_group = before["ret"]["DBSubnetGroups"][0]
        try:
            subnet_list = list()
            for each_subnet in old_db_subnet_group.get("Subnets"):
                subnet_list.append(each_subnet.get("SubnetIdentifier"))
            if set(subnets) != set(subnet_list):
                is_updated = True
            db_subnet_group_arn = old_db_subnet_group.get("DBSubnetGroupArn")
            ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=db_subnet_group_arn
            )
            if ret_tag["result"]:
                old_tag_list = ret_tag.get("ret").get("TagList")
                old_db_subnet_group["tags"] = old_tag_list
            else:
                result["comment"] = ret_tag["comment"]
                result["result"] = False
                return result
            if is_updated:
                update_ret = await hub.exec.boto3.client.rds.modify_db_subnet_group(
                    ctx,
                    DBSubnetGroupName=name,
                    DBSubnetGroupDescription=db_subnet_group_description,
                    SubnetIds=subnets,
                )
                if not update_ret["result"]:
                    result["comment"] = update_ret["comment"]
                    result["result"] = False
                    return result

            update_tags_ret = await hub.exec.aws.rds.tag.update_rds_tags(
                ctx=ctx,
                resource_arn=db_subnet_group_arn,
                old_tags=old_tag_list,
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["comment"] = update_tags_ret["comment"]
                result["result"] = False
                return result
            is_tags_updated = update_tags_ret["ret"]
            if is_updated or is_tags_updated:
                result["comment"] = f"Updated '{name}'"
            else:
                result["comment"] = f"'{name}' already exists"
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False
    else:
        try:
            ret = await hub.exec.boto3.client.rds.create_db_subnet_group(
                ctx,
                DBSubnetGroupName=name,
                DBSubnetGroupDescription=db_subnet_group_description,
                SubnetIds=subnets,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            db_subnet_group_arn = (
                ret["ret"].get("DBSubnetGroup").get("DBSubnetGroupArn")
            )
            result["comment"] = f"Created '{name}'"
            is_new = True
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
            return result

    try:
        if is_updated or is_new:
            after = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
                ctx, DBSubnetGroupName=name
            )
            new_db_subnet_group = after["ret"]["DBSubnetGroups"][0]
        else:
            new_db_subnet_group = copy.deepcopy(old_db_subnet_group)
        if is_tags_updated or is_new:
            ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=db_subnet_group_arn
            )
            new_tag_list = ret_tag.get("ret").get("TagList")
            new_db_subnet_group["tags"] = new_tag_list
        else:
            new_db_subnet_group["tags"] = old_tag_list

        result["old_state"] = old_db_subnet_group
        result["new_state"] = new_db_subnet_group
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a DB subnet group.  The specified database subnet group must not be associated with any DB instances.

    Args:
        name(Text, mandatory): A name, ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.rds.db_subnet_group.absent:
                - name: value
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    try:
        before = await hub.exec.boto3.client.rds.describe_db_subnet_groups(
            ctx, DBSubnetGroupName=name
        )
        if not before:
            result["comment"] = f"'{name}' already absent"
            return result
        elif ctx.get("test", False):
            result["comment"] = f"Would delete aws.rds.db_subnet_group {name}"
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.rds.delete_db_subnet_group(
                    ctx, DBSubnetGroupName=name
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = f"Deleted '{name}'"
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = f"{e.__class__.__name__}: {e}"

        result["old_state"] = before["ret"]["DBSubnetGroups"][0]
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of DBSubnetGroup descriptions. For an overview of CIDR ranges, go to the Wikipedia Tutorial.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.rds.db_subnet_group
    """

    result = {}
    try:
        ret = await hub.exec.boto3.client.rds.describe_db_subnet_groups(ctx)
        if not ret["result"]:
            hub.log.debug(f"Could not describe db_subnet_group {ret['comment']}")
            return {}

        for db_subnet_group in ret["ret"]["DBSubnetGroups"]:
            db_subnet_group_id = db_subnet_group.get("DBSubnetGroupName")
            describe_parameters = OrderedDict(
                {"DBSubnetGroupDescription": "db_subnet_group_description"}
            )
            new_db_subnet_group = []
            for parameter_old_key, parameter_new_key in describe_parameters.items():
                if db_subnet_group.get(parameter_old_key) is not None:
                    new_db_subnet_group.append(
                        {parameter_new_key: db_subnet_group.get(parameter_old_key)}
                    )

            if db_subnet_group.get("Subnets"):
                subnets = list()
                for each_subnet in db_subnet_group.get("Subnets"):
                    subnets.append(each_subnet.get("SubnetIdentifier"))

                new_db_subnet_group.append({"subnets": subnets})

            db_subnet_group_arn = db_subnet_group.get("DBSubnetGroupArn")
            ret_tag = await hub.exec.boto3.client.rds.list_tags_for_resource(
                ctx, ResourceName=db_subnet_group_arn
            )
            tag_list = ret_tag.get("ret").get("TagList")
            if tag_list:
                new_db_subnet_group.append({"tags": tag_list})
            result[db_subnet_group_id] = {
                "aws.rds.db_subnet_group.present": new_db_subnet_group
            }
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.associate_route_table
hub.exec.boto3.client.ec2.create_route_table
hub.exec.boto3.client.ec2.delete_route_table
hub.exec.boto3.client.ec2.describe_route_tables
hub.exec.boto3.client.ec2.disassociate_route_table
resource = hub.tool.boto3.resource.create(ctx, "ec2", "RouteTable", name)
hub.tool.boto3.resource.exec(resource, associate_with_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
"""
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.vpc.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    associations: List = None,
    routes: List = None,
    propagating_vgws: List = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a route table for the specified VPC. After you create a route table, you can add routes and associate
    the table with a subnet. For more information, see Route tables in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): A name, ID to identify the resource.
        vpc_id(Text): ID of the VPC.
        associations(list, optional): The associations between the route table and one or more subnets or a gateway.
            Defaults to None.
            * SubnetId (Text) -- The ID of the subnet. A subnet ID is not returned for an implicit association.
            * GatewayId (Text) -- The ID of the internet gateway or virtual private gateway.
        routes(list, optional): The routes in the route table.
            Defaults to None.
            DestinationCidrBlock (Text) -- The IPv4 CIDR block used for the destination match.
            DestinationIpv6CidrBlock (Text) -- The IPv6 CIDR block used for the destination match.
            DestinationPrefixListId (Text) -- The prefix of the Amazon Web Service.
            EgressOnlyInternetGatewayId (Text) -- The ID of the egress-only internet gateway.
            GatewayId (Text) -- The ID of a gateway attached to your VPC.
            InstanceId (Text) -- The ID of a NAT instance in your VPC.
            InstanceOwnerId (Text) -- The ID of Amazon Web Services account that owns the instance.
            NatGatewayId (Text) -- The ID of a NAT gateway.
            TransitGatewayId (Text) -- The ID of a transit gateway.
            LocalGatewayId (Text) -- The ID of the local gateway.
            CarrierGatewayId (Text) -- The ID of the carrier gateway.
            NetworkInterfaceId (Text) -- The ID of the network interface.
            VpcPeeringConnectionId (Text) -- The ID of a VPC peering connection.
            CoreNetworkArn (Text) -- The Amazon Resource Name (ARN) of the core network.
        propagating_vgws(list, optional): Any virtual private gateway (VGW) propagating routes.
            GatewayId (Text) -- The ID of the virtual private gateway.
        tags(List, optional): The tags to assign to the route_table. Defaults to None.
            * Key (string) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [route_table-resource-id]:
          aws.ec2.route_table.present:
          - vpc_id: 'string'
          - associations:
            - GatewayId: 'string'
            - SubnetId: 'string'
          - routes:
            - DestinationCidrBlock: 'string'
              GatewayId: 'string'
          - propagating_vgws:
             - 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            rtb-0944888360196462c:
              aws.ec2.route_table.present:
              - vpc_id: vpc-02850adfa9f6fc916
              - associations:
                - GatewayId: vgw-0334141b528f99316
                - SubnetId: subnet-0610f7c1a12a1af6a
              - routes:
                - DestinationCidrBlock: 198.31.0.0/16
                  GatewayId: local
              - propagating_vgws:
                 - vgw-0334141b528f99316
              - tags:
                - Key: Name
                  Value: route-table-association-test
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    created_route_table_id = name
    update_tags = False
    before = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, DryRun=False, RouteTableIds=[name]
    )

    if before and before["result"] and before["ret"].get("RouteTables"):
        route_table_resource = before["ret"].get("RouteTables")[0]
        update_ret = await hub.exec.aws.ec2.route_table.update_associations_routes(
            ctx=ctx,
            route_table_id=route_table_resource.get("RouteTableId"),
            old_routes=route_table_resource.get("Routes"),
            new_routes=routes,
            old_associations=route_table_resource.get("Associations"),
            new_associations=associations,
            old_propagating_vgws=route_table_resource.get("PropagatingVgws"),
            new_propagating_vgws=propagating_vgws,
        )
        update_comment = update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
        update_tags_comment = ""
        if result["result"] and tags is not None:
            # Update tags
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=route_table_resource.get("RouteTableId"),
                old_tags=route_table_resource.get("Tags"),
                new_tags=tags,
            )
            update_tags = True
            update_tags_comment = update_ret["comment"]
            if not update_ret["result"]:
                result["result"] = False
        result[
            "comment"
        ] = f"{name} is already present. {update_comment}. {update_tags_comment}"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_route_table(
                ctx,
                DryRun=False,
                TagSpecifications=[{"ResourceType": "route-table", "Tags": tags}]
                if tags
                else None,
                VpcId=vpc_id,
            )
            result["result"] = ret["ret"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            created_route_table_id = ret["ret"]["RouteTable"]["RouteTableId"]
            ret["comment"] = f"Created '{created_route_table_id}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    after = None
    if update_tags or created_route_table_id is not None:
        after = await hub.exec.boto3.client.ec2.describe_route_tables(
            ctx, DryRun=False, RouteTableIds=[created_route_table_id]
        )
    if before["result"] and before["ret"].get("RouteTables"):
        result["old_state"] = before["ret"]["RouteTables"][0]
    if after and after["result"] and after["ret"].get("RouteTables"):
        result["new_state"] = after["ret"]["RouteTables"][0]
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified route table. You must disassociate the route table from any subnets before you can delete
    it. You can't delete the main route table.

    Args:
        name(Text): A name or ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.route_table.absent:
                - name: value
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    route_table_id_list = [name]
    before = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, DryRun=False, RouteTableIds=route_table_id_list
    )
    if before and not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if before["result"] and before["ret"].get("RouteTables"):
        result["old_state"] = before["ret"]["RouteTables"][0]

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.ec2.route_table {name}"
        return result
    else:
        try:
            # we have to disassociate all subnets and gateways before we can delete route table
            if before["ret"] and before["ret"]["RouteTables"]:
                old_route_table = before["ret"]["RouteTables"][0]
                if old_route_table.get("Associations"):
                    for association in old_route_table.get("Associations"):
                        if association.get("Main") is not True:
                            ret = await hub.exec.boto3.client.ec2.disassociate_route_table(
                                ctx,
                                AssociationId=association.get(
                                    "RouteTableAssociationId"
                                ),
                            )
                            if not ret.get("result"):
                                result["comment"] = ret["comment"]
                                result["result"] = False
                                return result
            ret = await hub.exec.boto3.client.ec2.delete_route_table(
                ctx, RouteTableId=name
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your route tables. Each subnet in your VPC must be associated with a route table. If a
    subnet is not explicitly associated with any route table, it is implicitly associated with the main route table.
    This command does not return the subnet ID for implicit associations. For more information, see Route tables in
    the Amazon Virtual Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.route_table
    """

    result = {}

    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe route_table {ret['comment']}")
        return result
    describe_parameters = OrderedDict(
        {
            "VpcId": "vpc_id",
            "PropagatingVgws": "propagating_vgws",
            "Tags": "tags",
        }
    )
    for route_table in ret["ret"]["RouteTables"]:
        route_table_id = route_table.get("RouteTableId")

        new_route_table = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if route_table.get(parameter_old_key) is not None:
                new_route_table.append(
                    {parameter_new_key: route_table.get(parameter_old_key)}
                )
        # Add only required parameters in Describe
        if route_table.get("Routes"):
            routes_to_add = []
            for route_to_add in route_table.get("Routes"):
                new_route = route_to_add
                if route_to_add.get("State") == "active":
                    del new_route["State"]
                    del new_route["Origin"]
                    routes_to_add.append(new_route)
            new_route_table.append({"routes": routes_to_add})

        if route_table.get("Associations"):
            associations_to_add = []
            for association_to_add in route_table.get("Associations"):
                new_association = association_to_add
                if (
                    association_to_add.get("AssociationState").get("State")
                    == "associated"
                    and not new_association["Main"]
                ):
                    del new_association["AssociationState"]
                    del new_association["Main"]
                    associations_to_add.append(new_association)
            new_route_table.append({"associations": associations_to_add})
        result[route_table_id] = {"aws.ec2.route_table.present": new_route_table}
    return result



DESCRIBE_FUNCTIONS = ("describe", "get", "search", "list")

DELETE_FUNCTIONS = (
    "delete",
    "disassociate",
    "reject",
    "deallocate",
    "unassign",
    "deregister",
    "deprovision",
    "revoke",
    "release",
    "terminate",
    "cancel",
    "disable",
)

CREATE_FUNCTIONS = (
    "create",
    "associate",
    "accept",
    "allocate",
    "assign",
    "register",
    "provision",
    "authorize",
    "run",
    "enable",
    "upload",
    "put",
    "publish",
)

NAME_PARAMETER = {
    "default": None,
    "doc": "A name, ID to identify the resource",
    "param_type": "Text",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}


PRESENT_REQUEST_FORMAT = r"""
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    {{ function.hardcoded.resource_function_call }}
    before = {{ function.hardcoded.describe_function_call }}
    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} {name}"
            result["result"] = True
        return result

    if before:
        result["comment"] = f"'{name}' already exists"
        # TODO perform day-2 modifications as needed here
    else:
        try:
            ret = await {{ function.hardcoded.create_function }}(
                ctx,
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    {{ function.hardcoded.waiter_call }}
    # TODO perform other modifications as needed here
    ...

    try:
        after = {{ function.hardcoded.describe_function_call }}
        result["new_state"] = after
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result
"""

ABSENT_REQUEST_FORMAT = r"""
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    {{ function.hardcoded.resource_function_call }}
    before = {{ function.hardcoded.describe_function_call }}

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }} {name}"
        return result
    else:
        result["old_state"] = before
        try:
            ret = await {{ function.hardcoded.delete_function }}(
                ctx,
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    {{ function.hardcoded.waiter_call }}

    after = {{ function.hardcoded.describe_function_call }}
    result["changes"] = differ.deep_diff(before, after)
    return result
"""

DESCRIBE_REQUEST_FORMAT = r"""
    result = {}
    ret = await {{ function.hardcoded.list_function}}(ctx)

    if not ret["status"]:
        hub.log.debug(f"Could not describe {{ function.hardcoded.resource }} {ret['comment']}")
        return {}

    for {{ function.hardcoded.resource }} in ret["ret"]["{{ function.hardcoded.list_item }}"]:
        # Including fields to match the 'present' function parameters
        # TODO convert the dictionary values from string to object by removing the quotes.
        # TODO From 'resource[param]' to resource[param]
        new_{{ function.hardcoded.resource }} = {{ function.hardcoded.present_params }}
        result[{{ function.hardcoded.resource }}["{{ function.hardcoded.resource_id }}"]] = {"{{ function.ref }}.present": new_{{ function.hardcoded.resource }}}

    return result
"""



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.organizations.deregister_delegated_administrator
hub.exec.boto3.client.organizations.list_delegated_administrators
hub.exec.boto3.client.organizations.register_delegated_administrator
"""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub, ctx, name: str, org_unit_name: str, parent_id: str, tags: List = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an organizational unit (OU) within a root or parent OU. An OU is a container
    for accounts that enables you to organize your accounts to apply policies according
    to your business requirements. The number of levels deep that you can nest OUs is
    dependent upon the policy types enabled for that root. For service control policies,
    the limit is five.

    Args:
        hub:
        ctx:
        name(Text): A name or ID to identify the resource.
        parent_id(str): The unique identifier (ID) of the parent root or OU that you
                                  want to create the new OU in.
        org_unit_name(str): A friendly name to assign to the ou
        tags(list,Optional): A list of tags that you want to attach to the newly created OU


    Request Syntax:
        [org-id]:
          aws.organizations.organization_unit.present:
          - parent_id: 'string'
          - org_unit_name: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            org_is_present:
              aws.organizations.organization_unit.present:
                - parent_id: ou-01919
                - org_unit_name: ou-new
                - tags:
                    - Key: org
                      Value: guardrails
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    ou_id = name

    before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
        ctx, OrganizationalUnitId=ou_id
    )

    if ctx.get("test", False):
        if before:
            result[
                "comment"
            ] = f"Would update aws.organizations.organization_unit {name}."
            result["result"] = True
        else:
            result[
                "comment"
            ] = f"Would create aws.organizations.organization_unit {name}."
            result["result"] = True
        return result

    if not before:
        # organization unit does not exist, create
        hub.log.debug("Creating organization for this account")

        try:
            create_ret = (
                await hub.exec.boto3.client.organizations.create_organizational_unit(
                    ctx, ParentId=parent_id, Name=org_unit_name, Tags=tags
                )
            )

            result["result"] = create_ret["result"]

            if not result["result"]:
                result["comment"] = create_ret["comment"]
                return result
            ou_id = create_ret["ret"]["OrganizationalUnit"]["Id"]
            result["comment"] = f"Created aws.organizations.organization_unit {name}."
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
    else:
        # organizational_unit exists , update
        existing_ou_name = before["ret"]["OrganizationalUnit"]["Name"]

        before["ret"].pop("ResponseMetadata", None)

        old_tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=ou_id
        )

        if old_tags:
            before["ret"]["OrganizationalUnit"]["Tags"] = old_tags["ret"]["Tags"]

        result["old_state"] = before["ret"]["OrganizationalUnit"]

        try:

            if existing_ou_name != org_unit_name:

                update_ret = await hub.exec.boto3.client.organizations.update_organizational_unit(
                    ctx,
                    OrganizationalUnitId=ou_id,
                    Name=org_unit_name,
                )

                result["result"] = update_ret["result"]
                if not result["result"]:
                    result["comment"] = update_ret["comment"]
                    return result
                result[
                    "comment"
                ] = f"Updated aws.organizations.organization_unit {name}."

            if tags is not None and old_tags["result"]:
                update_tags_ret = (
                    await hub.exec.aws.organizations.organization.update_tags(
                        ctx, ou_id, old_tags["ret"].get("Tags", []), tags
                    )
                )
                if not update_tags_ret["result"]:
                    result["comment"] = "{} , {}".format(
                        result["comment"], str(update_tags_ret["comment"])
                    )
                    return result
                result["comment"] = "{} , {}".format(
                    result["comment"],
                    f"Updated tags on aws.organizations.organization_unit '{name}'. ",
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    try:
        after = await hub.exec.boto3.client.organizations.describe_organizational_unit(
            ctx, OrganizationalUnitId=ou_id
        )

        if after.get("ret"):
            after["ret"].pop("ResponseMetadata", None)
            updated_tag = (
                await hub.exec.boto3.client.organizations.list_tags_for_resource(
                    ctx, ResourceId=ou_id
                )
            )
            if updated_tag and updated_tag.get("ret"):
                after["ret"]["OrganizationalUnit"]["Tags"] = updated_tag["ret"].get(
                    "Tags", []
                )
            result["new_state"] = after["ret"]["OrganizationalUnit"]
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes an organizational unit (OU) from a root or another OU. You must first remove all accounts
    and child OUs from the OU that you want to delete.This operation can be called only from the
    organization's management account.

    Args:
        name(Text): An ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.organizations.organization_unit.absent:
                - name: ou-id1011
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)

    ou_id = name

    before = await hub.exec.boto3.client.organizations.describe_organizational_unit(
        ctx, OrganizationalUnitId=ou_id
    )

    if ctx.get("test", False):
        if before:
            result[
                "comment"
            ] = f"Would delete aws.organizations.organization_unit {name}."
            result["result"] = True
        else:
            result[
                "comment"
            ] = f"aws.organizations.organization_unit {name} already absent"
            result["result"] = True
        return result

    if not before:
        result["comment"] = f"aws.organizations.organization_unit {name} already absent"
        result["result"] = True
    else:
        before["ret"].pop("ResponseMetadata", None)

        result["old_state"] = before["ret"]["OrganizationalUnit"]
        try:
            ret = await hub.exec.boto3.client.organizations.delete_organizational_unit(
                ctx, OrganizationalUnitId=ou_id
            )

            if not ret:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"aws.organizations.organization_unit {name} deleted."
            result["result"] = ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Lists the organizational units (OUs) in a parent organizational unit or root.

    Idem does not support passing arguments to the describe function as of today , hence describe organizational unit
    will describe all the organizational unit one level under the root of the organization.
    In future if Idem starts supporting input arguments , we can pass any parent_id and this function will list
    all organizational units one level under the parent.


    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.organizations.organization_unit
    """

    result = {}

    list_roots_resp = await hub.exec.boto3.client.organizations.list_roots(ctx)

    if not (list_roots_resp and list_roots_resp["ret"].get("Roots")):
        return result

    parent_id = list_roots_resp["ret"]["Roots"][0]["Id"]

    if parent_id is None:
        return result

    org_units = (
        await hub.exec.boto3.client.organizations.list_organizational_units_for_parent(
            ctx, ParentId=parent_id
        )
    )

    if not org_units:
        hub.log.debug(
            f"Could not describe organization unit {org_units['comment']} for parent_id {parent_id}"
        )
        result[
            "comment"
        ] = f"Could not describe aws.organizations.organization_unit for parent_id {parent_id}"
        result["result"] = False
        return result

    organizational_units = org_units["ret"]["OrganizationalUnits"]

    for ou in organizational_units:
        translated_resource = []
        if ou.get("Name") is not None:
            translated_resource.append({"org_unit_name": ou.get("Name")})
        translated_resource.append({"parent_id": parent_id})
        tags = await hub.exec.boto3.client.organizations.list_tags_for_resource(
            ctx, ResourceId=ou["Id"]
        )
        if tags:
            translated_resource.append({"tags": tags["ret"].get("Tags", [])})
        result[ou["Id"]] = {
            "aws.organizations.organization_unit.present": translated_resource
        }

    return result



from typing import Any
from typing import Dict
from typing import List


async def update_cidr_blocks(
    hub,
    ctx,
    vpc_id: str,
    old_ipv4_cidr_blocks: List[Dict[str, Any]],
    old_ipv6_cidr_blocks: List[Dict[str, Any]],
    new_ipv4_cidr_blocks: List[Dict[str, Any]],
    new_ipv6_cidr_blocks: List[Dict[str, Any]],
):
    """
    Update associated cidr blocks of a vpc. This function compares the existing(old) cidr blocks and the
    new cidr blocks. Cidr blocks that are in the new_cidr_blocks but not in the old_cidr_blocks will be associated to
    vpc. Cidr blocks that are in the old_cidr_blocks but not in the new_cidr_blocks will be disassociated from vpc.

    Args:
        hub:
        ctx:
        vpc_id: The AWS resource id of the existing vpc
        old_ipv4_cidr_blocks: The ipv4 cidr blocks on the existing vpc
        old_ipv6_cidr_blocks: The ipv6 cidr blocks on the existing vpc
        new_ipv4_cidr_blocks: The expected ipv4 cidr blocks on the existing vpc
        new_ipv6_cidr_blocks: The expected ipv6 cidr blocks on the existing vpc

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    # If a block is None, we'll skip updating such cidr block.
    if new_ipv4_cidr_blocks is None:
        new_ipv4_cidr_blocks = old_ipv4_cidr_blocks
    if new_ipv6_cidr_blocks is None:
        new_ipv6_cidr_blocks = old_ipv6_cidr_blocks
    result = dict(comment="", result=True, ret=None)
    ipv4_cidr_block_map = dict()
    for cidr_block in old_ipv4_cidr_blocks:
        if cidr_block.get("CidrBlockState").get("State") == "associated":
            ipv4_cidr_block_map[cidr_block.get("CidrBlock")] = cidr_block
    ipv6_cidr_block_map = dict()
    for cidr_block in old_ipv6_cidr_blocks:
        if cidr_block.get("Ipv6CidrBlockState").get("State") == "associated":
            ipv6_cidr_block_map[cidr_block.get("Ipv6CidrBlock")] = cidr_block
    cidr_blocks_to_add = list()
    cidr_blocks_to_remove = list()
    for cidr_block_set in new_ipv4_cidr_blocks:
        if cidr_block_set.get("CidrBlock") not in ipv4_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    cidr_block_set, "ipv4"
                )
            )
        else:
            del ipv4_cidr_block_map[cidr_block_set.get("CidrBlock")]
    for ipv6_cidr_block_set in new_ipv6_cidr_blocks:
        if ipv6_cidr_block_set.get("Ipv6CidrBlock") not in ipv6_cidr_block_map:
            cidr_blocks_to_add.append(
                hub.tool.aws.network_utils.generate_cidr_request_payload_for_vpc(
                    ipv6_cidr_block_set, "ipv6"
                )
            )
        else:
            del ipv6_cidr_block_map[ipv6_cidr_block_set.get("Ipv6CidrBlock")]
    cidr_blocks_to_remove = (
        cidr_blocks_to_remove
        + [
            cidr_block.get("AssociationId")
            for cidr_block in ipv4_cidr_block_map.values()
        ]
        + [
            cidr_block.get("AssociationId")
            for cidr_block in ipv6_cidr_block_map.values()
        ]
    )
    for request_payload in cidr_blocks_to_add:
        ret = await hub.exec.boto3.client.ec2.associate_vpc_cidr_block(
            ctx, VpcId=vpc_id, **request_payload
        )
        if not ret.get("result"):
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
    for association_id in cidr_blocks_to_remove:
        ret = await hub.exec.boto3.client.ec2.disassociate_vpc_cidr_block(
            ctx, AssociationId=association_id
        )
        if not ret.get("result"):
            result["comment"] = ret["comment"]
            result["result"] = False
            return result
    result[
        "comment"
    ] = f"Update tags: Add [{cidr_blocks_to_add}] Remove [{cidr_blocks_to_remove}]"
    return result
